************************************************************************
*  Calcul d'exposants de singularites en elasticite anisotrope.
*  Yvon Lafranche, Universite de Rennes 1.
*  Voir documentation a l'adresse :
*        http://perso.univ-rennes1.fr/yvon.lafranche/exsiel/
************************************************************************
************************************************************************
* DECHOL : decomposition de Cholesky (A = LLt)
* ----------------------------------------------------------------------
* BERSTO : resolution de l'equation P(x) = 0 par la methode de Bairstow
* CALPQ  : calcul des coefficients p et q du polynome (x**2 + p*x + q)
* DIVPD2 : division d'un polynome par (x**2 + p*x + q)
* RESPD2 : resolution du polynome (x**2 + p*x + q)
* ----------------------------------------------------------------------
* RACPCR : determination d'une racine d'un polynome a coefficients reels
*          par la methode de Newton
* VALPCR : evaluation d'un polynome a coefficients reels (sch de Horner)
* ----------------------------------------------------------------------
* INIEPS : initialisation des epsilons (commun defini dans epsilon.inc)
* RSIPRE : detection automatique de la precision utilisee
************************************************************************
      subroutine DECHOL(A,nlamax,n,L,nllmax,eps,codret)
      integer nlamax, n, nllmax, codret
      real A(nlamax,*), L(nllmax,*), eps
*  Variables locales :
      integer i, j, k
      real s
*       -------------------------------------------------------------
*       Decomposition de Cholesky de la matrice A : A = LLt
*         Arguments d'entree  : A,nlamax,n,nllmax,eps
*         Arguments de sortie : L,codret
*       A      : matrice donnee d'ordre n symetrique definie positive
*       nlamax : premiere dimension de A specifiee dans l'ordre
*                de declaration du module appelant
*       n      : ordre de la matrice
*       L      : matrice triangulaire inferieure resultat de la
*                decomposition (si A n'est pas reutilisee, A et L
*                peuvent correspondre au meme emplacement memoire)
*       nllmax : premiere dimension de L specifiee dans l'ordre de
*                declaration du module appelant
*       eps    : sert a determiner la positivite de la matrice A
*       codret : code de retour valant 0 si RAS, 1 si la matrice A
*                n'est pas definie positive.
*
*       --> seuls sont utilises les elements des matrices
*            triangulaires inferieures
*       --> utiliser ensuite DRCHOL pour resoudre un systeme lineaire
*       -------------------------------------------------------------
      do 30 i=1,n
       do 20 j=i,n
        s = A(j,i)
        do 10 k=1,i-1
   10    s = s - L(i,k)*L(j,k)
   20   L(j,i) = s
       if (L(i,i) .lt. 0.) goto 40
       L(i,i) = sqrt(L(i,i))
       if (L(i,i) .lt. eps) goto 40
       s = L(i,i)
       do 30 j=i+1,n
   30   L(j,i) = L(j,i)/s
      codret = 0
      return
*
   40 codret = 1      
      end
************************************************************************
************************************************************************
*               FICHIER D'INCLUSION 'epsilon.inc'
*
*    Definition des epsilons necessaires au calcul :
*     eps0   : seuil au dessous duquel une valeur est consideree nulle
*     epsmac : epsilon machine (plus petit e>0 tq 1+e = 1)
*              Sur SUN, epsmac = EPMRSP(r), r real simple ou double prec
*     epsrel : seuil d'erreur relative (Delta r / r < epsrel)
*    Ces valeurs doivent etre initialisees par INIEPS.
*
**      real eps0, epsmac, epsrel
**      common /epsiln/ eps0, epsmac, epsrel
************************************************************************
************************************************************************
      subroutine BERSTO(n,a,ap,b,z,codret)
      integer n, codret
      real a(0:n), ap(0:n), b(0:n)
      complex z(n)
*  Variables locales
      integer nc, i, nbrac
      real p, q
*       -------------------------------------------------------------
*       Resolution de l'equation P(x) = 0 par la methode de Bairstow.
*       P est un polynome de degre n a coefficients reels.
*       Arguments d'entree  : n,a
*       Arguments de sortie : z,codret
*       Tableaux de travail : ap,b
*        n : degre du polynome
*        a : coefficients du polynome (a(i) coeff de x**i, i=0,n)
*        z : racines trouvees (tableau de complexes)
*        codret : code de retour valant
*           0 si RAS
*           1 si la factorisation du polynome a echoue
*           2 si le nombre de racines trouvees est different de n
*       -------------------------------------------------------------
      codret = 0
      nbrac = 1
*       Sauvegarde des coefficients du polynome et du degre
      do 10 i=0,n
   10  ap(i) = a(i)
      nc = n
*
   20 if (nc .ge. 3) then
       if (ap(0) .eq. 0.) then
*        0 est racine
        z(nbrac) = cmplx(0.)
        nbrac = nbrac + 1
        nc = nc - 1
        do 30 i=0,nc
   30    ap(i) = ap(i+1)
       else 
        call DIVPD2(nc,ap,b,p,q,codret)
        if (codret .ne. 0) goto 50
        call RESPD2(p,q,z(nbrac))
        nbrac = nbrac + 2
        nc = nc - 2
        do 40 i=0,nc
   40    ap(i) = b(i)
       endif 
       goto 20
      elseif (nc .eq. 2) then
       p = ap(1)/ap(2)
       q = ap(0)/ap(2)
       call RESPD2(p,q,z(nbrac))
       nbrac = nbrac + 2
      elseif (nc .eq. 1) then
       z(nbrac) = cmplx(-ap(0)/ap(1))
       nbrac = nbrac + 1
      endif
*
      if (nbrac-1 .ne. n) codret = 2
   50 return
      end
************************************************************************
      subroutine CALPQ(n,a,b,p,q,codret)
      integer n, codret
      real a(0:n), b(0:n), p, q
*  Variables locales
      integer niter, niterx, i
      real r,s, c0,c1,ct, drp,drq,dsp,dsq, det, delp,delq
      parameter (niterx = 50)
*  Declarations generales et communs
      include 'epsilon.inc'
*       ----------------------------------------------------------
*       Calcul des coefficients p et q du polynome (x**2 + p*x + q)
*       qui divise le polynome P de coefficients (a(i),i=0,n).
*       P est un polynome de degre n a coefficients reels.
*       Arguments d'entree  : n,a,p,q
*       Arguments de sortie : b,p,q,codret
*        n : degre du polynome
*        a : coefficients du polynome (a(i) coeff de x**i, i=0,n)
*        b : coefficients du polynome quotient (b(i),i=0,n-2)
*        p,q : en entree, point de depart de la methode de Newton
*              en sortie, coefficients du diviseur x**2 + p*x + q
*        codret : code de retour valant 0 si RAS, 1 si la methode
*                 de Newton n'a pu converger.
*       ----------------------------------------------------------
*      Schema de la methode de Newton : si X=(p,q),
*        X(k+1) = Xk - Inverse(Ak) (f(Xk))   ou Ak = f'(Xk)
*
      codret = 0
      niter = 0
   10 niter = niter + 1
*       Calcul des b(i) (division par x**2 + p*x + q)
      b(n-1) = 0.
      b(n-2) = a(n)
      do 20 i=n-1,2,-1
   20  b(i-2) = a(i) - b(i-1)*p - b(i)*q
*       Evaluation de la fonction f(Xk) = (r(p,q), s(p,q))
      r = a(1) - b(0)*p - b(1)*q
      s = a(0) - b(0)*q
*       Calcul de c0 et c1
      c0 = 0.
      c1 = 0.
      do 30 i=n-2,1,-1
       ct = c1
       c1 = c0
   30  c0 = -(b(i) + c1*p + ct*q)
*       Calcul des coefficients de l'inverse de Ak = 1/det ( dsq -drq)
*                                                          (-dsp  drp)
      dsp = -c0*q
      dsq = -c1*q -b(0)
      drp = -c0*p + dsq
      drq = c0
      det = drp*dsq - dsp*drq
      if (abs(det) .lt. eps0) goto 40
      delp = ( dsq*r - drq*s) / det
      delq = (-dsp*r + drp*s) / det
*       Calcul de la nouvelle solution
      p = p - delp
      q = q - delq
*       Test d'arret
      ct = abs(delp)+abs(delq)
      if (ct.gt.epsmac .and. ct/(abs(p)+abs(q)).gt.epsrel
     & .and. niter.lt.niterx) goto 10
      if (niter.eq.niterx) goto 40
      return
*
   40 codret = 1
      end
************************************************************************
      subroutine DIVPD2(n,a,b,p,q,codret)
      integer n, codret
      real a(0:n), b(0:n), p, q
*  Variables locales
      integer i,j,k, nbv
      parameter (nbv = 10)
      real rmin, rmax, h, x1,x2
*       -----------------------------------------------------------
*       Division du polynome P de coefficients (a(i),i=0,n) par le
*       polynome (x**2 + p*x + q).
*       P est un polynome de degre n a coefficients reels.
*       Arguments d'entree  : n,a
*       Arguments de sortie : b,p,q,codret
*        n : degre du polynome
*        a : coefficients du polynome (a(i) coeff de x**i, i=0,n)
*        b : coefficients du polynome quotient (b(i),i=0,n-2)
*        p,q : coefficients du diviseur x**2 + p*x + q
*        codret : code de retour valant 0 si RAS, 1 si la methode
*                 de Newton n'a pu converger.
*       -----------------------------------------------------------
      codret = 0
*      Les racines de P sont dans le disque de centre O et de rayon
*      rmax, et a l'exterieur du disque de centre O et de rayon rmin.
*      Determination des rayons des disques.
      rmin = abs(a(1))
      do 10 i=2,n-1
   10  rmin = max(rmin,abs(a(i)))
      rmax = rmin
      rmin = max(rmin,abs(a(n)))
      rmax = max(rmax,abs(a(0)))
      rmin = 1. / (1. + rmin/abs(a(0)))
      rmax = 1. + rmax/abs(a(n))
      h = (rmax-rmin) / real(nbv-1)
*      Balayage en (p,q), point de depart de la methode de Newton
      do 20 i=-1,1,2
       x1 = (rmin-h)*i
       do 20 j=-1,1,2
        x2 = (rmin-h)*j
        do 20 k=1,nbv
         x1 = x1 + i*h
         x2 = x2 + j*h
         q = x1*x2
         p = -x1-x2
         call CALPQ(n,a,b,p,q,codret)
         if (codret .eq. 0) goto 30
   20 continue   
   30 end
************************************************************************
      subroutine RESPD2(p,q,z)
      real p, q
      complex z(2)
*  Variables locales
      real r, discri
*       ----------------------------------------------------------
*       Resolution du polynome a coefficients reels x**2 + p*x + q
*       Arguments d'entree  : p,q
*       Arguments de sortie : z
*        p,q : coefficients du polynome x**2 + p*x + q
*        z   : racines du polynome
*       ----------------------------------------------------------
      discri = p*p - 4*q
      if (discri .lt. 0.) then
       discri = sqrt(-discri)*0.5
       r = -p*0.5
       z(1) = cmplx(r,discri)
       z(2) = cmplx(r,-discri)
      else
       r = 0.5*(-p + sign(1.,-p)*sqrt(discri))
       z(1) = cmplx(r)
       z(2) = cmplx(q/r)
      endif
      end
************************************************************************
************************************************************************
      subroutine RACPCR(n,a,ap,z,codret)
      integer n, codret
      real a(0:n), ap(0:n)
      complex z, z0
*  Variables locales
      integer niter, niterx, i, nm1
      complex fz, fpz, dz
      parameter (niterx = 50)
*  Declarations generales et communs
      include 'epsilon.inc'
*  Point d'entree
      complex VALPCR
*       ---------------------------------------------------------
*       Determination d'une racine z d'un polynome a coefficients
*       reels par la methode de Newton.
*       Arguments d'entree  : n,a,z
*       Arguments de sortie : z
*       Tableau de travail  : ap
*        n : degre du polynome
*        a : coefficients du polynome (a(i) coeff de x**i, i=0,n)
*        z : en entree, point de depart de la methode de Newton
*            en sortie, racine recherchee (voir codret)
*        codret : code de retour valant 0 si RAS, 1 si la methode
*                 de Newton n'a pu converger ; dans ce cas la
*                 valeur de z est inchangee.
*       --> L'appel a VALPCR provoque (sur SUN) un message du
*       type Underflow (IEEE floating-point arithmetic exception)
*       lorsque l'operande z comporte une composante "petite".
*       ---------------------------------------------------------
      codret = 0
      z0 = z
*      Calcul des coefficients du polynome derive
      nm1 = n-1
      do 10 i=1,n
   10  ap(i-1) = i*a(i)
*
      niter = 0
   20 niter = niter + 1
      fpz = VALPCR(nm1,ap,z)
      if (abs(fpz) .lt. eps0) goto 30
      fz = VALPCR(n,a,z)
      dz = fz/fpz
      z = z - dz
      if (abs(fz).gt.eps0 .and. abs(dz).gt.epsmac
     & .and. niter.lt.niterx) goto 20
      return
*
   30 codret = 1
      z = z0
      end
************************************************************************
      complex function VALPCR(n,a,z)
      integer n
      real a(0:n)  
      complex z
*  Variables locales
      complex r
      integer i
*       ---------------------------------------------------------
*       Evaluation au point z du polynome P a coefficients reels
*       par l'algorithnme de Horner.
*       Arguments d'entree  : n,a,z
*       Arguments de sortie : VALPCR
*        n : degre du polynome
*        a : coefficients du polynome (a(i) coeff de x**i, i=0,n)
*        z : point d'evaluation
*       ---------------------------------------------------------
      r = cmplx(a(n))
      do 10 i=n-1,0,-1
  10   r = r*z + cmplx(a(i))
      VALPCR = r
      end
************************************************************************
************************************************************************
      subroutine INIEPS(type)
      integer type
*  Variables locales
      logical sp
*  Declarations generales et communs
      include 'epsilon.inc'
*  Point d'entree
      logical RSIPRE
*    -------------------------------------------------------------------
*    Definition des epsilons necessaires au calcul memorises dans le
*    commun defini dans 'epsilon.inc' :
*     eps0   : seuil au dessous duquel une valeur est consideree nulle
*     epsmac : epsilon machine (plus petit e>0 tq 1+e = 1)
*              Sur SUN, epsmac = EPMRSP(r), r real simple ou double prec
*     epsrel : seuil d'erreur relative (Delta r / r < epsrel)
*    L'initialisation est faite selon la valeur d'entree de type :
*      . type = 1 : en simple precision
*      . type = 2 : en double precision
*      . type = autre valeur : autodetection de la precision utilisee
*                              et initialisation en consequence
*    Arguments d'entree : type
*    -------------------------------------------------------------------
      sp = type .eq. 1
      if (type.ne.1 .and. type.ne.2) then
*  Autodetection de la precision
       sp = RSIPRE()
      endif
*
      if (sp) then
*  Simple precision
       eps0 = 1.e-20
       epsmac = 1.1920928955078125e-7
       epsrel = 5.e-5
      else
*  Double precision (option de compilation -r8 sur SUN)
       eps0 = 1.d-30
       epsmac = 2.220446049250313080847263336181640625d-16
       epsrel = 5.d-14
      endif
      end
************************************************************************
      logical function RSIPRE()
*  Variable locales
*    On peut aussi prendre t de type character*4 et vref='zzzz' par
*    exemple, mais l'equivalence (character-autre type) est refusee
*    par certains compilateurs.
      integer*4 t(2), vref
      parameter (vref = -1)
      real r
      equivalence (r,t)
*       ---------------------------------------------------------------
*       Teste si le type real est code sur 4 ou 8 octets.
*       Par defaut, real est code sur 4 octets.  Cependant, si le
*       programme a ete compile avec une option demandant une precision
*       superieure (typiquement option -r8 sur SUN), le codage est
*       realise sur  8 octets.
*       Le sous-programme renvoie VRAI pour un codage sur 4 octets,
*       (Real SImple PREcision), FAUX sinon.
*       ---------------------------------------------------------------
      t(1) = vref
      t(2) = vref
      r = 0.
      RSIPRE = t(2).eq.vref
      end
**************************** Fin du fichier ****************************
